// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Collections;

namespace MySql.Data.MySqlClient
{
	/// <summary>
	/// Summary description for CommandResult.
	/// </summary>
	internal class CommandResult
	{
		private Driver				driver;

		private long				affectedRows;
		private long				fieldCount;
		private int					warningCount;
		private long				lastInsertId;
		private ServerStatusFlags	serverStatus;
		private string				serverMessage;

		private int					fieldsRead;
		private bool				readSchema;
		private bool				readRows;
		private bool				isBinary;
		private MySqlField[]		fields;

		public CommandResult(Packet p, Driver d, bool binary)
		{
			driver = d;
			isBinary = binary;

			GetNextResultSet(p);
		}

		public CommandResult( Driver d )
		{
			driver = d;
		}

		#region Properties

		public MySqlField[] Fields 
		{
			get { return fields; }
		}

		public bool IsResultSet 
		{
			get { return fieldCount > 0; }
		}

		public long LastInsertId 
		{
			get { return lastInsertId; }
			set { lastInsertId = value; }
		}

		public long FieldCount 
		{
			get { return fieldCount; }
			set { fieldCount = value; }
		}

		public long AffectedRows
		{
			get { return affectedRows; }
			set { affectedRows = value; }
		}

		public int WarningCount 
		{
			get { return warningCount; }
			set { warningCount = value; }
		}

		public string ServerMessage 
		{ 
			get { return serverMessage; }
			set { serverMessage = value; }
		}

		public ServerStatusFlags ServerStatus 
		{
			get { return serverStatus; }
			set { serverStatus = value; }
		}
		#endregion

		public bool GetNextResultSet(Packet packet)
		{
			fieldCount = affectedRows = 0;
			readSchema = readRows = false;

			// if we are pre-4.1 and are asking for 2nd and later resultsets, then just return
			// false since that isn't possible.
//			if (packet == null)
//			{
//				if (! driver.Version.isAtLeast(4,1,0)) return false;
//				if (serverStatus
//				return false;

			

			while ( packet != null  || 
				(serverStatus & (ServerStatusFlags.MoreResults | ServerStatusFlags.AnotherQuery )) != 0)
			{
				// if we are getting subsequent queries, then the driver *MIGHT* have to bump
				// it's packet sequence number (compression)
				//if ( (serverStatus & ServerStatusFlags.AnotherQuery) != 0)
					//driver.BumpCommandSequence();

				if (packet == null)
					packet = driver.ReadPacket();

				fieldCount = packet.ReadPackedLong();
				if (fieldCount > 0) return true;

				ReadStatus(packet);
				packet = null;
			} 
			return false;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public bool Load()
		{
			// read in field defs
			fields = new MySqlField[ fieldCount ];
			for (int i=0; i < fieldCount; i++)
				fields[i] = GetField();

			// read in end of meta data packet
			Packet last = driver.ReadPacket();
			ReadStatus(last);
			readSchema = true;

			// see if we have data in this resultset
			Packet data = driver.PeekPacket();
			return ! data.IsLastPacket();
		}

		public MySqlField GetField()
		{
			bool is41 = driver.Version.isAtLeast(4,1,0);
			fieldsRead++;
			if (is41) 
				return new MySqlField41( driver );
			else
				return new MySqlField( driver );
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public bool ReadDataRow()
		{
			if ( readRows ) return false;

			Packet packet = driver.ReadPacket();
			if (packet.IsLastPacket())
			{
				ReadStatus(packet);
				readRows = true;
				return false;
			}

			if (isBinary)
				SetBinaryFieldData(packet);
			else
				SetFieldData(packet);
			return true;
		}

		/// <summary>
		/// 
		/// </summary>
		private void SetFieldData(Packet packet)
		{
			for (int col=0; col < fieldCount; col++)
			{
				long len = (long)packet.GetFieldLength();
				fields[col].SetTextData( packet.GetBuffer(), packet.Position, 
					len, driver.Version );
				if (len != -1)
					packet.Position += len;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		private void SetBinaryFieldData(Packet packet)
		{
			//TODO:  what is this byte?
			packet.ReadByte();
			byte[] nullMapBytes = new byte[ (fieldCount+7)/8 ];
			packet.Read( nullMapBytes, 0, nullMapBytes.Length );
			BitArray nullMap = new BitArray( nullMapBytes );

			for (int col=0; col < fieldCount; col++)
			{
				MySqlField41 field = (fields[col] as MySqlField41);
				if (nullMap[col+2])
					field.SetBinaryData( null, driver.Version );
				else 
				{
					field.SetBinaryData( packet, driver.Version );
					//TODO: remove the setBinaryData returning length
				}
			}
		}

		/// <summary>
		/// Read status info
		/// </summary>
		/// <param name="p"></param>
		private void ReadStatus(Packet p)
		{
			p.Position = 0;
			byte key = (byte)p.ReadByte();
			if (key == 0)
				ReadOkPacket(p);
			else if (p.Length < 9 && key == 254 && driver.Version.isAtLeast(4,1,0))
			{
				warningCount = p.ReadInteger(2);
				serverStatus = (ServerStatusFlags)p.ReadInteger(2);
			}
		}

		private void ReadOkPacket(Packet p)
		{
			if (!driver.Version.isAtLeast(4,1,0))
			{
				affectedRows = p.ReadPackedInteger();
				lastInsertId = p.ReadPackedInteger();
			}
			else 
			{
				affectedRows = p.ReadPackedLong();
				lastInsertId = p.ReadPackedLong();
				serverStatus = (ServerStatusFlags)p.ReadInteger(2);
				warningCount = p.ReadInteger(2);
				if (p.HasMoreData)
					serverMessage = p.ReadLenString();
			}
		}

		/// <summary>
		/// 
		/// </summary>
		public void Consume()
		{
			Packet p;

			// if we are not a resultset, then we are done
			if (! this.IsResultSet) return;

			if (! readSchema)
			{
				do 
				{
					p = driver.ReadPacket();
				} while (! p.IsLastPacket());
				ReadStatus(p);
			}

			if (! readRows)
			{
				do 
				{
					p = driver.ReadPacket();
				} while (! p.IsLastPacket());
				ReadStatus(p);
			}
		}
	}
}
