// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Collections;
using MySql.Data.Common;

namespace MySql.Data.MySqlClient
{
	/// <summary>
	/// 
	/// </summary>
	internal sealed class InternalConnection
	{
		MySqlConnectionString	settings;
		Driver					driver;
		CommandResult			lastResult;
		DateTime				createTime;

		public InternalConnection( MySqlConnectionString connectString )
		{
			settings = connectString;
		}

		#region Properties
		public MySqlConnectionString Settings 
		{
			get { return settings; }
			set { settings = value; driver.Settings = value; }
		}

		internal Driver Driver 
		{
			get { return driver; }
		}

		/// <summary>
		/// Returns true if this connection can handle batch SQL natively
		/// This means MySQL 4.1.1 or later.
		/// </summary>
		public bool SupportsBatch 
		{
			get { return (driver.Flags & ClientFlags.MULTI_STATEMENTS) != 0;}
		}

		#endregion

		#region Methods

		public bool IsAlive() 
		{
			try 
			{
				Packet p = driver.Send( DBCmd.PING, (byte[])null );
				// we don't care about the result.  The fact that it responded is enough
				return true;
			}
			catch (Exception)
			{
				return false;
			}
		}

		public bool IsTooOld() 
		{
			TimeSpan ts = DateTime.Now.Subtract( createTime );
			if (ts.Seconds > settings.ConnectionLifetime)
				return true;
			return false;
		}


		/// <summary>
		/// I don't like this setup but can't think of a better way of doing
		/// right now.
		/// </summary>
		/// <param name="connection"></param>
		public void Configure(MySqlConnection connection)
		{
			// load server properties
			Hashtable props = new Hashtable();
			MySqlCommand cmd = new MySqlCommand("SHOW VARIABLES", connection);

			try 
			{
				MySqlDataReader reader = cmd.ExecuteReader();
				while (reader.Read()) 
					props.Add( reader.GetValue(0), reader.GetValue(1) );
				reader.Close();
			}
			catch (Exception ex)
			{
				Logger.LogException( ex );
				throw ex;
			}

#if AUTHENTICATED
			string licenseType = props["license"];
			if (licenseType == null || licenseType.Length == 0 || 
				licenseType != "commercial") 
				throw new MySqlException( "This client library licensed only for use with commercially-licensed MySQL servers." );
#endif

			if (props.Contains( "max_allowed_packet" ))
				driver.MaxPacketSize = Convert.ToInt64(props["max_allowed_packet"]);

			string charSet = null;
			if (! driver.Version.isAtLeast(4,1,0))
			{
				if (props.Contains( "character_set" ))
					charSet = props["character_set"].ToString();
			}
			else 
			{

				charSet = "utf8";
				cmd.CommandText = "SET character_set_client = utf8";
				cmd.ExecuteNonQuery();
			}

			if (charSet != null)
				driver.Encoding = CharSetMap.GetEncoding( charSet );
			else
				throw new MySqlException("Unknown text encoding.  Aborting!");
		}

		/// <summary>
		/// Open this internal connection object
		/// </summary>
		public void Open() 
		{
			driver = new Driver();
			driver.Open( settings );

			createTime = DateTime.Now;
		}

		public void Close() 
		{
			driver.Close();
		}


		/// <summary>
		/// Sets the current database for the this connection
		/// </summary>
		/// <param name="dbName"></param>
		public void SetDatabase( string dbName )
		{
			byte[] dbNameBytes = driver.Encoding.GetBytes( dbName );
			CommandResult result = ExecuteCmd( DBCmd.INIT_DB, dbNameBytes, true );
		}

		public void Reset()
		{
			driver.ResetUser();
		}

		/// <summary>
		/// Executes a single command against the currently connected server
		/// </summary>
		/// <param name="cmd">The command to execute</param>
		/// <param name="buffer">The text of the command</param>
		/// <param name="consume">True to consume all output, false otherwise</param>
		/// <returns>CommandResult object containing the results of the command</returns>
		/// <remarks>
		/// The MySqlConnectionInternal object keeps a reference to the last
		/// CommandReesult object returned.
		/// </remarks>
		public CommandResult ExecuteCmd( DBCmd cmd, byte[] buffer, bool consume )
		{
			// if we are logging, then write out the command
			if (settings.Logging)
				Logger.LogCommand( cmd, driver.GetString(buffer) );


			// execute the command
			Packet packet = driver.Send( cmd, buffer );
			CommandResult result = new CommandResult( packet, driver, false );
			lastResult = result;

			// if we are supposed to consume the results of this command, then do so now
			if (consume)
				result.Consume();

			return result;
		}

		public CommandResult ExecuteStatement( byte[] bytes )
		{
			Packet result = driver.Send( DBCmd.EXECUTE, bytes );
			return new CommandResult( result, driver, true );
		}

		/// <summary>
		/// Prepare a SQL statement for later execution
		/// </summary>
		/// <param name="sql">SQL to prepare</param>
		/// <returns>PreparedStatement object containing the metadata for the prepared statement</returns>
		public PreparedStatement Prepare( string sql )
		{
			byte[] sqlBuffer = driver.GetBytes( sql );
			Packet packet = driver.Send( DBCmd.PREPARE, sqlBuffer );

			int marker = packet.ReadByte();
			if (marker != 0)
				throw new MySqlException("Expected prepared statement marker");

			int statementId = packet.ReadInteger(4);
			int numCols = packet.ReadInteger(2);
			int numParms = packet.ReadInteger(2);

			return new PreparedStatement( this, statementId, numCols, numParms );
		}


		#endregion

	}
}
