// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Net;
using System.Net.Sockets;
using System.IO;
using ICSharpCode.SharpZipLib.Zip.Compression;
using ICSharpCode.SharpZipLib.Zip.Compression.Streams;
using System.Security.Cryptography;
using MySql.Data.Common;
using System.Collections;
using System.Text;

namespace MySql.Data.MySqlClient
{
	/// <summary>
	/// Summary description for Driver.
	/// </summary>
	internal class Driver
	{
		protected const int HEADER_LEN = 4;
		protected const int MIN_COMPRESS_LENGTH = 50;

		protected int					maxSinglePacket = 255 * 255 * 255;
		protected Stream				stream;
		protected BufferedStream		writer;
		protected Encoding				encoding;
		protected byte					packetSeq;
		protected byte					cmdPacketSeq;
		protected long					maxPacketSize;
		protected DBVersion				serverVersion;
		protected bool					isOpen;
		protected string				versionString;
		protected Packet				peekedPacket;

		protected int					protocol;
		protected uint					threadID;
		protected String				encryptionSeed;
		protected ClientFlags			serverCaps;
		protected bool					useCompression = false;
		protected int					serverStatus;
		protected ClientFlags			connectionFlags;
		protected MySqlConnectionString	connectionString;

		public Driver()
		{
			packetSeq = 0;
			cmdPacketSeq = 0;
			isOpen = false;
			encoding = System.Text.Encoding.Default;
			maxPacketSize = 1047552;
		}

		public ClientFlags Flags
		{
			get { return connectionFlags; }
		}

		public Encoding Encoding 
		{
			get 
			{ 
				return encoding; 
			}
			set { encoding = value; }
		}

		public long MaxPacketSize 
		{
			get { return maxPacketSize; }
			set { maxPacketSize = value; }
		}

		public string VersionString 
		{
			get { return versionString; }
		}

		public MySqlConnectionString Settings 
		{
			get { return connectionString; }
			set { connectionString = value; }
		}

		public DBVersion Version 
		{
			get { return serverVersion; }
		}

		public string GetString( byte[] stringBuffer )
		{
			return encoding.GetString( stringBuffer );
		}

		public byte[] GetBytes( string s )
		{
			return encoding.GetBytes( s );
		}

		public void Open( MySqlConnectionString settings )
		{
			// save this for later use
			connectionString = settings;

			// connect to one of our specified hosts
			try 
			{
				if (settings.Protocol == ConnectionProtocol.SharedMemory)
				{
					SharedMemoryStream str = new SharedMemoryStream( settings.SharedMemoryName );
					str.Open( settings.ConnectionTimeout );
					stream = str;
				}
				else 
				{
					string pipeName = settings.PipeName;
					if (settings.Protocol != ConnectionProtocol.NamedPipe)
						pipeName = null;
					StreamCreator sc = new StreamCreator( settings.Server, settings.Port, pipeName );
					stream = sc.GetStream( settings.ConnectionTimeout );
				}
			}
			catch (Exception ex)
			{
				throw new MySqlException("Unable to connect to any of the specified MySQL hosts", ex);
			}


			if (stream == null) 
				throw new MySqlException("Unable to connect to any of the specified MySQL hosts");

			writer = new BufferedStream( stream );
			// read off the welcome packet and parse out it's values
			Packet packet = ReadPacket();
			protocol = packet.ReadByte();
			versionString = packet.ReadString();
			serverVersion = DBVersion.Parse( versionString );
			threadID = (uint)packet.ReadInteger(4);
			encryptionSeed = packet.ReadString();

			// starting with 4.0.8, maxSinglePacket should be 0xffffff
			if ( serverVersion.isAtLeast(4,0,8))
				maxSinglePacket = (256*256*256)-1;

			// read in Server capabilities if they are provided
			serverCaps = 0;
			if (packet.HasMoreData)
				serverCaps = (ClientFlags)packet.ReadInteger(2);

			// based on our settings, set our connection flags
			SetConnectionFlags();

			Packet authPacket = CreatePacket(null);
			authPacket.WriteInteger( (int)connectionFlags, serverVersion.isAtLeast(4,1,0) ? 4 : 2 );
			authPacket.WriteInteger( maxSinglePacket, serverVersion.isAtLeast(4,1,0) ? 4 : 3 );


			// 4.1.1 included some new server status info
			if (serverVersion.isAtLeast(4,1,1))
			{
				/* New protocol with 16 bytes to describe server characteristics */
				int serverLanguage = packet.ReadInteger(2);
				serverStatus = packet.ReadInteger(2);
				packet.Position += 12;

				string seedPart2 = packet.ReadString();
				encryptionSeed += seedPart2;

				authPacket.WriteByte( 8 );					// connect with latin1
				authPacket.Write( new byte[23], 0, 23 );	// reserved space
			}

			Authenticate( authPacket );

			// if we are using compression, then we use our CompressedStream class
			// to hide the ugliness of managing the compression
			if ((connectionFlags & ClientFlags.COMPRESS) != 0)
			{
				useCompression = true;
				stream = new CompressedStream( stream );
				writer = new BufferedStream( stream );
			}

			isOpen = true;
		}

		/// <summary>
		/// Return the appropriate set of connection flags for our
		/// server capabilities and our user requested options.
		/// </summary>
		private void SetConnectionFlags()
		{
			ClientFlags flags = ClientFlags.FOUND_ROWS;

			if (serverVersion.isAtLeast(4,1,1))
			{
				flags |= ClientFlags.PROTOCOL_41;
				// Need this to get server status values
				flags |= ClientFlags.TRANSACTIONS;

				// user allows/disallows batch statements
				if (connectionString.AllowBatch)
					flags |= ClientFlags.MULTI_STATEMENTS;

				// We always allow multiple result sets
				flags |= ClientFlags.MULTI_RESULTS;
			}
			else if (serverVersion.isAtLeast( 4, 1, 0 ))
				flags |= ClientFlags.RESERVED;
			
			// if the server allows it, tell it that we want long column info
			if ((serverCaps & ClientFlags.LONG_FLAG) != 0)
				flags |= ClientFlags.LONG_FLAG;

			// if the server supports it and it was requested, then turn on compression
			if ((serverCaps & ClientFlags.COMPRESS) != 0 && connectionString.UseCompression)
				flags |= ClientFlags.COMPRESS;

			if ( protocol > 9 )
				flags |= ClientFlags.LONG_PASSWORD; // for long passwords
			else 
				flags &= ~ClientFlags.LONG_PASSWORD;

			// allow load data local infile
			flags |= ClientFlags.LOCAL_FILES;

			// if the server allows it and a database was specified, then indicate
			// that we will connect with a database name
			if ((serverCaps & ClientFlags.CONNECT_WITH_DB) != 0 && 
				connectionString.Database != null && connectionString.Database.Length > 0)
				flags |= ClientFlags.CONNECT_WITH_DB;

			// if the server is requesting a secure connection, then we oblige
			if ((serverCaps & ClientFlags.SECURE_CONNECTION) != 0)
				flags |= ClientFlags.SECURE_CONNECTION;

			connectionFlags = flags;
		}


		/// <summary>
		/// Perform an authentication against a 4.1.1 server
		/// </summary>
		/// <param name="packet">The partially complete authentication packet</param>
		private void Authenticate411( Packet packet )
		{
			if ( (connectionFlags & ClientFlags.SECURE_CONNECTION) == 0)
				AuthenticateOld( packet );

			packet.Write( Crypt.Get411Password( connectionString.Password, this.encryptionSeed ));
			if ( (connectionFlags & ClientFlags.CONNECT_WITH_DB) != 0)
				packet.WriteString( connectionString.Database, encoding );
			SendPacket( packet );
			Packet p = ReadPacket();

			// this result means the server wants us to send the password using
			// old encryption
			if (p.IsLastPacket())
			{
				p = CreatePacket(null);
				p.WriteString( Crypt.EncryptPassword( 
					connectionString.Password, this.encryptionSeed.Substring(0,8), true ),
					encoding );
				SendPacket( p );
				p = ReadPacket();
			}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="packet"></param>
		private void Authenticate410( Packet packet )
		{
			if (connectionString.Password == null || connectionString.Password.Length == 0)
				packet.WriteByte(0);
			else
				packet.WriteString( "xxxxxxxx", encoding );
			if ( (connectionFlags & ClientFlags.CONNECT_WITH_DB) != 0)
				packet.WriteString( connectionString.Database, encoding );
			SendPacket( packet );

			Packet p = ReadPacket();
			if (p.Length == 1) return;

			byte[] replyBytes = p.GetBuffer();
			
			p = CreatePacket(null);
			if (replyBytes[0] != '*')
				p.Write( Crypt.Get410Password( connectionString.Password, replyBytes ));
			else
				p.Write( Crypt.GetOld410Password( connectionString.Password, replyBytes ) );

			SendPacket(p);
			p = ReadPacket();
		}

		private void AuthenticateOld( Packet packet )
		{
			packet.WriteString( Crypt.EncryptPassword( 
				connectionString.Password, encryptionSeed, protocol > 9), 
				encoding );
			if ( (connectionFlags & ClientFlags.CONNECT_WITH_DB) != 0)
				packet.WriteString( connectionString.Database, encoding );
			SendPacket( packet );
			Packet p = ReadPacket();
		}

		public void Authenticate(Packet packet)
		{
			// write the user id to the auth packet
			packet.WriteString( connectionString.UserId, encoding );

			if (serverVersion.isAtLeast(4,1,1))
				Authenticate411( packet );
			else if (serverVersion.isAtLeast(4,1,0))
				Authenticate410( packet );
			else
				AuthenticateOld( packet );
		}

		public void ResetUser()
		{
			this.packetSeq = 0;
			this.cmdPacketSeq = 0;
			Packet p = CreatePacket(null);
			p.WriteByte( (byte)DBCmd.CHANGE_USER );
			Authenticate( p );
		}

		public Packet CreatePacket( byte[] buf )
		{
			if (buf == null)
				return new Packet( serverVersion.isAtLeast(3, 22, 5) );
			return new Packet( buf, serverVersion.isAtLeast(3, 22, 5 ));
		}
		
		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public Packet PeekPacket()
		{
			if (peekedPacket != null)
				return peekedPacket;

			peekedPacket = ReadPacket();
			return peekedPacket;
		}

		/// <summary>
		/// ReadBuffer continuously loops until it has read the entire
		/// requested data
		/// </summary>
		/// <param name="buf">Buffer to read data into</param>
		/// <param name="offset">Offset to place the data</param>
		/// <param name="length">Number of bytes to read</param>
		private void ReadBuffer( byte[] buf, int offset, int length )
		{
			while (length > 0)
			{
				int amountRead = stream.Read( buf, offset, length );
				if (amountRead == 0)
					throw new MySqlException("Unexpected end of data encountered");
				length -= amountRead;
				offset += amountRead;
			}
		}

		internal void BumpCommandSequence() 
		{
			if (useCompression)
				packetSeq = (byte)(cmdPacketSeq++);
		}

		private Packet ReadPacketFromServer()
		{
			int len = stream.ReadByte() + (stream.ReadByte() << 8) +
				(stream.ReadByte() << 16);
			byte seq = (byte)stream.ReadByte();
			byte[] buf = new byte[ len ];
			ReadBuffer( buf, 0, len );

//			if (seq != packetSeq) 
//				throw new MySqlException("Unknown transmission status: sequence out of order");
			packetSeq++;

			Packet p = CreatePacket(buf);
			p.Encoding = this.Encoding;
			if (p.Length == maxSinglePacket && serverVersion.isAtLeast(4,0,0)) 
				p.Append( ReadPacketFromServer() );
			return p;
		}

		/// <summary>
		/// Reads a single packet off the stream
		/// </summary>
		/// <returns></returns>
		public Packet ReadPacket()
		{
			// if we have peeked at a packet, then return it
			if (peekedPacket != null)
			{
				Packet packet = peekedPacket;
				peekedPacket = null;
				return packet;
			}

			Packet p = ReadPacketFromServer();

			// if this is an error packet, then throw the exception
			if (p[0] == 0xff)
			{
				p.ReadByte();
				int errorCode = (int)p.ReadInteger(2);
				string msg = p.ReadString();
				throw new MySqlException( msg, errorCode );
			}
			
			return p;
		}

		protected MemoryStream CompressBuffer(byte[] buf, int index, int length)
		{

			if (length < MIN_COMPRESS_LENGTH) return null;

			MemoryStream ms = new MemoryStream(buf.Length);
			DeflaterOutputStream dos = new DeflaterOutputStream(ms);

			dos.WriteByte( (byte)(length & 0xff ));
			dos.WriteByte( (byte)((length >> 8) & 0xff ));
			dos.WriteByte( (byte)((length >> 16) & 0xff ));
			dos.WriteByte( 0 );

			dos.Write( buf, index, length );
			dos.Finish();
			if (ms.Length > length+4) return null;
			return ms;
		}

		private void WriteInteger( int v, int numbytes )
		{
			int val = v;

			if (numbytes < 1 || numbytes > 4) 
				throw new ArgumentOutOfRangeException("Wrong byte count for WriteInteger");

			for (int x=0; x < numbytes; x++)
			{
				writer.WriteByte( (byte)(val&0xff) );
				val >>= 8;
			}
		}

		/// <summary>
		/// Send a buffer to the server in a compressed form
		/// </summary>
		/// <param name="buf">Byte buffer to send</param>
		/// <param name="index">Location in buffer to start sending</param>
		/// <param name="length">Amount of data to send</param>
		protected void SendCompressedBuffer(byte[] buf, int index, int length)
		{
			MemoryStream compressed_bytes = CompressBuffer(buf, index, length);
			int comp_len = compressed_bytes == null ? length+HEADER_LEN : (int)compressed_bytes.Length;
			int ucomp_len = compressed_bytes == null ? 0 : length+HEADER_LEN;

			WriteInteger( comp_len, 3 );
			writer.WriteByte( packetSeq++ );
			WriteInteger( ucomp_len, 3 );
			if (compressed_bytes != null)
				writer.Write( compressed_bytes.GetBuffer(), 0, (int)compressed_bytes.Length );
			else 
			{
				WriteInteger( length, 3 );	
				writer.WriteByte( 0 );
				writer.Write( buf, index, length );
			}
			stream.Flush();
		}

		/// <summary>
		/// Send a single packet to the server.
		/// </summary>
		/// <param name="packet">Packet to send to the server</param>
		/// <remarks>This method will send a single packet to the server
		/// possibly breaking the packet up into smaller packets that are
		/// smaller than max_allowed_packet.  This method will always send at
		/// least one packet to the server</remarks>
		protected void SendPacket(Packet packet)
		{
			byte[]	buf = packet.GetBuffer();
			int		len = packet.Length;
			int		index = 0;
			bool	oneSent = false;

			// make sure we are not trying to send too much
			if (packet.Length > maxPacketSize && maxPacketSize > 0)
				throw new MySqlException("Packet size too large.  This MySQL server cannot accept rows larger than " + maxPacketSize + " bytes.");

			try 
			{
				while (len > 0 || ! oneSent) 
				{
					int lenToSend = Math.Min( len, maxSinglePacket );

					// send the data
					if (useCompression)
						SendCompressedBuffer( buf, index, lenToSend );
					else 
					{
						WriteInteger( lenToSend, 3 );
						writer.WriteByte( packetSeq++ );
						writer.Write( buf, index, lenToSend );
						writer.Flush();
					}

					len -= lenToSend;
					index += lenToSend;
					oneSent = true;
				}
				writer.Flush();
			}
			catch (Exception ex)
			{
				Console.WriteLine( ex.Message );
			}
		}


		public void Close() 
		{
			if (stream != null)
				stream.Close();
		}


		/// <summary>
		/// Sends the specified command to the database
		/// </summary>
		/// <param name="command">Command to execute</param>
		/// <param name="text">Text attribute of command</param>
		/// <returns>Result packet returned from database server</returns>
		public Packet Send( DBCmd command, String text ) 
		{
			return Send( command, this.Encoding.GetBytes( text ) );
		}

		public Packet Send( DBCmd cmd, byte[] bytes )
		{
			//			string s = Encoding.GetString( bytes );

			Packet packet = CreatePacket(null);
			packetSeq = 0;
			cmdPacketSeq = 0;
			packet.WriteByte( (byte)cmd );
			if (bytes != null)
				packet.Write( bytes, 0, bytes.Length );

			SendPacket( packet );
			packet = ReadPacket();

			// first check to see if this is a LOAD DATA LOCAL callback
			// if so, send the file and then read the results
			long fieldcount = packet.ReadPackedLong();
			if (fieldcount == Packet.NULL_LEN)
			{
				string filename = packet.ReadString();
				SendFileToServer( filename );
				packet = ReadPacket();
			}
			else
				packet.Position = 0;
			return packet;
		}

		public CommandResult SendQuery( byte[] bytes )
		{
			Packet p = Send( DBCmd.QUERY, bytes );
			return new CommandResult( p, this, false);
		}

		/// <summary>
		/// Sends the specified file to the server. 
		/// This supports the LOAD DATA LOCAL INFILE
		/// </summary>
		/// <param name="filename"></param>
		private void SendFileToServer( string filename )
		{
			Packet		p = CreatePacket(null);
			byte[]		buffer = new byte[4092];
			FileStream	fs = null;
			try 
			{
				fs = new FileStream( filename, FileMode.Open );
				int count = fs.Read( buffer, 0, buffer.Length );
				while (count != 0) 
				{
					if ((p.Length + count) > maxSinglePacket)
					{
						SendPacket( p );
						p.Clear();
					}
					p.Write( buffer, 0, count );
					count = fs.Read( buffer, 0, buffer.Length );
				}
				fs.Close();

				// send any remaining data
				if (p.Length > 0) 
				{
					SendPacket(p);
					p.Clear();
				}
			}
			catch (Exception ex)
			{
				throw new MySqlException("Error during LOAD DATA LOCAL INFILE", ex);
			}
			finally 
			{
				if (fs != null)
					fs.Close();
				// empty packet signals end of file
				p.Clear();
				SendPacket(p);
			}
		}

	}
}
