// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using System.Data.SqlTypes;
using System.Globalization;
using System.Text;
using MySql.Data.Common;
using MySql.Data.Types;

namespace MySql.Data.MySqlClient
{
	internal enum ColumnFlags : int
	{
		NOT_NULL		= 1,
		PRIMARY_KEY		= 2,
		UNIQUE_KEY		= 4,
		MULTIPLE_KEY	= 8,
		BLOB			= 16,
		UNSIGNED		= 32,
		ZERO_FILL		= 64,
		BINARY			= 128,
		ENUM			= 256,
		AUTO_INCREMENT	= 512,
		TIMESTAMP		= 1024,
		SET				= 2048,
		NUMBER			= 32768
	};
	
	/// <summary>
	/// Summary description for Field.
	/// </summary>
	internal class MySqlField 
	{
		// public fields
		public		string		CatalogName;
		public		int			ColumnLength;
		public		string		ColumnName;
		public		string		TableName;
		public		string		RealTableName;
		public		string		DatabaseName;

		// protected fields
		protected	MySqlValue	colValue;
		protected	ColumnFlags	colFlags;
		protected	Encoding	encoding;
		protected	int			charSetIndex;

		protected	byte[]		buffer;
		protected	long		bufIndex;
		protected	long		bufLength;

		public MySqlField() 
		{
		}

		public MySqlField(Driver driver)
		{
			this.encoding = driver.Encoding;

			bool scaleNeedsBump = !driver.Version.isAtLeast(3,23,15) && driver.Version.isAtLeast(3,23,0);
			Packet packet = driver.ReadPacket();

			TableName = packet.ReadLenString();
			ColumnName = packet.ReadLenString();
			ColumnLength = packet.ReadNBytes();
			MySqlDbType type = (MySqlDbType)packet.ReadNBytes();
			packet.ReadByte();					// This is currently 2

			if ((driver.Flags & ClientFlags.LONG_FLAG) != 0)
				Flags = (ColumnFlags)packet.ReadInteger(2);
			else 
				Flags = (ColumnFlags)packet.ReadByte();

			byte scale = (byte)packet.ReadByte();

			if (scaleNeedsBump)
				scale++;

			InitValue( type, scale );
		}

		protected void InitValue(MySqlDbType type, byte scale) 
		{
			// create our MySqlValue object
			colValue = MySqlValue.GetMySqlValue( type, IsUnsigned, IsBinary );
			if (colValue is MySqlBinary)
				(colValue as MySqlBinary).IsBinary = IsBinary;

			if (colValue is MySqlDecimal)
			{
				(colValue as MySqlDecimal).Precision = (byte)ColumnLength;
				(colValue as MySqlDecimal).Scale = scale;
			}
		}

		public byte[] Buffer 
		{
			get { return buffer; }
		}

		public long	BufferIndex 
		{
			get { return bufIndex; }
		}

		public long BufferLength 
		{
			get { return bufLength; }
		}

		public int CharactetSetIndex 
		{
			get { return charSetIndex; }
			set 
			{ 
				if (value != charSetIndex)
					encoding = CharSetMap.GetEncoding( value );
				charSetIndex = value; 
			}
		}

		/// <summary>
		/// CopyBuffer makes a copy of the byte buffer given to us while
		/// the rowset was being read
		/// </summary>
		private void CopyBuffer()
		{
			byte[] newbuf = new byte[ bufLength ];
			Array.Copy( buffer, (int)bufIndex, newbuf, (int)0, (int)bufLength );
			bufIndex = 0;
			buffer = newbuf;
			colValue.ValueAsObject = buffer;
		}

		public MySqlValue MySqlValue 
		{
			get { return colValue; }
		}

		/// <summary>
		/// GetValue returns an object that represents the value of this field.
		/// </summary>
		/// <returns></returns>
		public object Value
		{	
			get 
			{
				// if our value is a byte buffer and we are using only 
				// a part of that buffer, then we need to make a copy
				if (colValue is MySqlBinary && (colValue as MySqlBinary).IsBinary &&
					(bufIndex > 0 || bufLength < buffer.Length))
					CopyBuffer();
				return colValue.ValueAsObject;
			}
		}

		public MySqlDbType	Type 
		{
			get { return colValue.MySqlDbType; }
		}

		public int Precision
		{
			get 
			{
				if (colValue is MySqlDecimal)
					return (colValue as MySqlDecimal).Precision;
				return -1;
			}

		}

		public int Scale
		{
			get 
			{
				if (colValue is MySqlDecimal)
					return (colValue as MySqlDecimal).Scale;
				return -1;
			}
		}

		public ColumnFlags Flags 
		{ 
			get { return colFlags; }
			set { colFlags = value; }
		}

		public bool IsAutoIncrement
		{
			get { return (colFlags & ColumnFlags.AUTO_INCREMENT) > 0; }
		}

		public bool IsNumeric
		{
			get { return (colFlags & ColumnFlags.NUMBER) > 0; }
		}

		public bool AllowsNull
		{
			get { return (colFlags & ColumnFlags.NOT_NULL) == 0; }
		}

		public bool IsUnique
		{
			get { return (colFlags & ColumnFlags.UNIQUE_KEY) > 0; }
		}

		public bool IsPrimaryKey
		{
			get { return (colFlags & ColumnFlags.PRIMARY_KEY) > 0; }
		}

		public bool IsBlob
		{
			get { return (colFlags & ColumnFlags.BLOB) > 0; }
		}

		public bool IsBinary
		{
			get { return (colFlags & ColumnFlags.BINARY) > 0; }
		}

		public bool IsUnsigned
		{
			get { return (colFlags & ColumnFlags.UNSIGNED) > 0; }
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="buf"></param>
		/// <param name="index"></param>
		/// <param name="len"></param>
		/// <param name="version"></param>
		public virtual void SetTextData( byte[] buf, long index, long len, DBVersion version )
		{
			if (len == -1)
			{
				colValue.IsNull = true;
				buffer = null;
				return;
			}

			buffer = buf;
			bufIndex = index;
			bufLength = len;

			bool is41 = version.isAtLeast(4,1,0);
			string sValue = encoding.GetString( buf, (int)index, (int)len );

			if (colValue is MySqlDateTime)
				(colValue as MySqlDateTime).ParseMySql( sValue, is41 );
			else if (colValue is MySqlTimeSpan)
				(colValue as MySqlTimeSpan).ParseMySql( sValue, is41 );
			else if (colValue is MySqlBinary) 
			{
				if (IsBinary) 
				{
					MySqlBinary binary = (colValue as MySqlBinary);
					binary.Value = buffer;
					binary.Index = (ulong)index;
					binary.Length = (ulong)len;
				}
				else
					(colValue as MySqlBinary).Value = sValue;
			}
			else
				colValue.Parse( sValue );
		}

		public string GetTypeName() 
		{
			return colValue.GetMySqlTypeName();
		}

		public Type GetFieldType() 
		{
			return colValue.SystemType;
		}

		public DbType GetDbType() 
		{
			return colValue.DbType;
		}
	}

}
