// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using MySql.Data.MySqlClient;

namespace MySql.Data.Types
{
	/// <summary>
	/// 
	/// </summary>
	public abstract class MySqlValue
	{
		/// <summary>The generic dbtype of this value</summary>
		protected DbType		dbType;

		/// <summary>The specific MySQL db type</summary>
		protected MySqlDbType	mySqlDbType;

		/// <summary>The MySQL specific typename of this value</summary>
		protected string		mySqlTypeName;

		/// <summary>The system type represented by this value</summary>
		protected Type			classType;

		/// <summary>Is this value null</summary>
		protected bool			isNull;

		/// <summary>
		/// 
		/// </summary>
		public MySqlValue() 
		{
			isNull = false;
		}

		// abstract methods
		internal abstract void	SerializeBinary( PacketWriter writer );
		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public abstract byte[] ToBytes();
		/// <summary>
		/// 
		/// </summary>
		/// <param name="s"></param>
		public abstract void Parse(string s);
		internal abstract long SetBinaryData( Packet packet );
		internal abstract string GetMySqlTypeName();
		internal abstract object ValueAsObject 
		{
			get; set;
		}
		internal abstract Type SystemType 
		{
			get;
		}

		/// <summary>
		/// 
		/// </summary>
		public bool IsNull 
		{
			get { return isNull; }
			set { isNull = value; }
		}

		internal virtual DbType DbType 
		{
			get { return dbType; }
		}

		internal MySqlDbType MySqlDbType 
		{
			get { return mySqlDbType; }
		}

		/// <summary>Returns a string representation of this value</summary>
		public override string ToString() 
		{
			return ValueAsObject.ToString();
		}

		internal static MySqlValue GetMySqlValueFromValue( object theValue ) 
		{
			return GetMySqlValueFromType( theValue.GetType() );
		}

		internal static MySqlValue GetMySqlValueFromDbType( DbType dbType ) 
		{
			switch (dbType) 
			{
				case DbType.Guid:
				case DbType.AnsiString:
				case DbType.AnsiStringFixedLength:
				case DbType.String:
				case DbType.StringFixedLength: return new MySqlString();
				case DbType.Boolean:
				case DbType.Byte: return new MySqlUByte();
				case DbType.SByte: return new MySqlByte();
				case DbType.Date:
				case DbType.DateTime: return new MySqlDateTime();
				case DbType.Time: return new MySqlTimeSpan();
				case DbType.Single: return new MySqlSingle();
				case DbType.Double: return new MySqlDouble();
				case DbType.Int16: return new MySqlInt16();
				case DbType.UInt16: return new MySqlUInt16();
				case DbType.Int32: return new MySqlInt32();
				case DbType.UInt32: return new MySqlUInt32();
				case DbType.Int64: return new MySqlInt64();
				case DbType.UInt64: return new MySqlUInt64();
				case DbType.Decimal:
				case DbType.Currency: return new MySqlDecimal();
				case DbType.Object:
				case DbType.VarNumeric: 
				case DbType.Binary: 
				default: return new MySqlBinary();
			}
		}

		internal static MySqlValue GetMySqlValueFromType( Type type )
		{
			switch (Type.GetTypeCode( type )) 
			{
				case TypeCode.SByte: return new MySqlByte();
				case TypeCode.Byte: return new MySqlByte();
				case TypeCode.Int16: return new MySqlInt16();
				case TypeCode.UInt16: return new MySqlUInt16();
				case TypeCode.Int32: return new MySqlInt32();
				case TypeCode.UInt32: return new MySqlUInt32();
				case TypeCode.Int64: return new MySqlInt64();
				case TypeCode.UInt64: return new MySqlUInt64();
				case TypeCode.DateTime: return new MySqlDateTime();
				case TypeCode.String: return new MySqlString();
			}

			if (type == typeof(TimeSpan)) return new MySqlTimeSpan();
			return new MySqlBinary();
		}

		internal static MySqlValue GetMySqlValue( MySqlDbType type, bool unsigned, bool binary )
		{
			switch (type) 
			{
				case MySqlDbType.Year:
				case MySqlDbType.Byte: 
					if (unsigned) return new MySqlUByte(type);
					return new MySqlByte();

				case MySqlDbType.Int16: 
					if (unsigned) return new MySqlUInt16();
					return new MySqlInt16();

				case MySqlDbType.Int24:
				case MySqlDbType.Int32: 
					if (unsigned) return new MySqlUInt32(type);
					return new MySqlInt32(type);

				case MySqlDbType.UInt16:
					return new MySqlUInt16( MySqlDbType.Int16 );
				case MySqlDbType.UInt32:
					return new MySqlUInt32( MySqlDbType.Int32 );
				case MySqlDbType.UInt64:
					return new MySqlUInt64( MySqlDbType.Int64 );

				case MySqlDbType.Int64: 
					if (unsigned) return new MySqlUInt64();
					return new MySqlInt64();

				case MySqlDbType.Time:
					return new MySqlTimeSpan();

				case MySqlDbType.Date:
				case MySqlDbType.Datetime:
				case MySqlDbType.Newdate:
				case MySqlDbType.Timestamp: return new MySqlDateTime(type);

				case MySqlDbType.Decimal: return new MySqlDecimal();

				case MySqlDbType.Float: return new MySqlSingle();

				case MySqlDbType.Double: return new MySqlDouble();

				case MySqlDbType.Set:
				case MySqlDbType.Enum:
				case MySqlDbType.String:
				case MySqlDbType.VarChar: return new MySqlString(type);

				case MySqlDbType.Blob:
				case MySqlDbType.MediumBlob:
				case MySqlDbType.LongBlob:
				default:
					if (binary) return new MySqlBinary(type);
					return new MySqlString( type );
			}
		}

	}

}
