// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using System.Text;
using System.ComponentModel;
using System.ComponentModel.Design.Serialization;
using System.Reflection;
using MySql.Data.Types;

namespace MySql.Data.MySqlClient
{
	/// <summary>
	/// Represents a parameter to a <see cref="MySqlCommand"/>, and optionally, its mapping to <see cref="DataSet"/> columns. This class cannot be inherited.
	/// </summary>
	[TypeConverter(typeof(MySqlParameter.MySqlParameterConverter))]
	public sealed class MySqlParameter : MarshalByRefObject, IDataParameter, IDbDataParameter, ICloneable
	{
		private object				oValue;
		private MySqlValue			paramValue;
		private ParameterDirection	direction = ParameterDirection.Input;
		private bool				isNullable  = false;
		private string				paramName;
		private string				sourceColumn;
		private DataRowVersion		sourceVersion = DataRowVersion.Current;
		private int					size;

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the MySqlParameter class.
		/// </summary>
		public MySqlParameter()
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="MySqlParameter"/> class with the parameter name and a value of the new MySqlParameter.
		/// </summary>
		/// <param name="parameterName">The name of the parameter to map. </param>
		/// <param name="value">An <see cref="Object"/> that is the value of the <see cref="MySqlParameter"/>. </param>
		public MySqlParameter(string parameterName, object value)
		{
			ParameterName = parameterName;
			Value = value;
			paramValue = MySqlValue.GetMySqlValueFromValue(value);

			if (value is Byte[])
				size = (value as Byte[]).Length;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="MySqlParameter"/> class with the parameter name and the data type.
		/// </summary>
		/// <param name="parameterName">The name of the parameter to map. </param>
		/// <param name="dbType">One of the <see cref="MySqlDbType"/> values. </param>
		public MySqlParameter( string parameterName, MySqlDbType dbType)
		{
			ParameterName = parameterName;
			paramValue = MySqlValue.GetMySqlValue( dbType, false, true );
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="MySqlParameter"/> class with the parameter name, the <see cref="MySqlDbType"/>, and the size.
		/// </summary>
		/// <param name="parameterName">The name of the parameter to map. </param>
		/// <param name="dbType">One of the <see cref="MySqlDbType"/> values. </param>
		/// <param name="size">The length of the parameter. </param>
		public MySqlParameter( string parameterName, MySqlDbType dbType, int size )
		{
			ParameterName = parameterName;
			paramValue = MySqlValue.GetMySqlValue( dbType, false, true );
			this.size = size;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="MySqlParameter"/> class with the parameter name, the <see cref="MySqlDbType"/>, the size, and the source column name.
		/// </summary>
		/// <param name="parameterName">The name of the parameter to map. </param>
		/// <param name="dbType">One of the <see cref="MySqlDbType"/> values. </param>
		/// <param name="size">The length of the parameter. </param>
		/// <param name="sourceColumn">The name of the source column. </param>
		public MySqlParameter( string parameterName, MySqlDbType dbType, int size, string sourceColumn )
		{
			ParameterName = parameterName;
			paramValue = MySqlValue.GetMySqlValue( dbType, false, true );
			this.size = size;
			this.direction = ParameterDirection.Input;
			this.sourceColumn = sourceColumn;
			this.sourceVersion = DataRowVersion.Current;
			Value = null;
		}

		internal MySqlParameter(string name, MySqlDbType type, ParameterDirection dir, string col, DataRowVersion ver, object val)
		{
			if (direction != ParameterDirection.Input)
				throw new ArgumentException("Only input parameters are supported by MySql");
			paramValue = MySqlValue.GetMySqlValue( type, false, true );
			direction = dir;
			ParameterName = name;
			sourceColumn = col;
			sourceVersion = ver;
			Value = val;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="MySqlParameter"/> class with the parameter name, the type of the parameter, the size of the parameter, a <see cref="ParameterDirection"/>, the precision of the parameter, the scale of the parameter, the source column, a <see cref="DataRowVersion"/> to use, and the value of the parameter.
		/// </summary>
		/// <param name="parameterName">The name of the parameter to map. </param>
		/// <param name="dbType">One of the <see cref="MySqlDbType"/> values. </param>
		/// <param name="size">The length of the parameter. </param>
		/// <param name="direction">One of the <see cref="ParameterDirection"/> values. </param>
		/// <param name="isNullable">true if the value of the field can be null, otherwise false. </param>
		/// <param name="precision">The total number of digits to the left and right of the decimal point to which <see cref="MySqlParameter.Value"/> is resolved.</param>
		/// <param name="scale">The total number of decimal places to which <see cref="MySqlParameter.Value"/> is resolved. </param>
		/// <param name="sourceColumn">The name of the source column. </param>
		/// <param name="sourceVersion">One of the <see cref="DataRowVersion"/> values. </param>
		/// <param name="value">An <see cref="Object"/> that is the value of the <see cref="MySqlParameter"/>. </param>
		/// <exception cref="ArgumentException"/>
		public MySqlParameter( string parameterName, MySqlDbType dbType, int size, ParameterDirection direction,
			bool isNullable, byte precision, byte scale, string sourceColumn, DataRowVersion sourceVersion,
			object value)
		{
			if (direction != ParameterDirection.Input)
				throw new ArgumentException("Only input parameters are supported by MySql");

			ParameterName = parameterName;
			paramValue = MySqlValue.GetMySqlValue( dbType, false, true );
			this.size = size;
			this.direction = direction;
			if (paramValue is MySqlDecimal) 
			{
				(paramValue as MySqlDecimal).Precision = precision;
				(paramValue as MySqlDecimal).Scale = scale;
			}
			this.sourceColumn = sourceColumn;
			this.sourceVersion = sourceVersion;
			Value = value;
		}
		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the <see cref="DbType"/> of the parameter.
		/// </summary>
		public DbType DbType 
		{
			get 
			{ 
				return paramValue.DbType;
			}
			set 
			{ 
				if (value == paramValue.DbType) return;
				paramValue = MySqlValue.GetMySqlValueFromDbType( value );
			}
		}

		/// <summary>
		/// Gets or sets a value indicating whether the parameter is input-only, output-only, bidirectional, or a stored procedure return value parameter.
		/// As of MySql version 4.1 and earlier, input-only is the only valid choice.
		/// </summary>
		[Category("Data")]
		public ParameterDirection Direction 
		{
			get { return direction; }
			set { direction = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether the parameter accepts null values.
		/// </summary>
		[Browsable(false)]
		public Boolean IsNullable 
		{
			get { return isNullable; }
		}

		/// <summary>
		/// Gets or sets the MySqlDbType of the parameter.
		/// </summary>
		[Category("Data")]
		public MySqlDbType MySqlDbType 
		{
			get  
			{ 
				return paramValue.MySqlDbType;
			}
			set  
			{
				if (paramValue != null && value == paramValue.MySqlDbType) return;
				paramValue = MySqlValue.GetMySqlValue( value, false, true );
			}
		}

		/// <summary>
		/// Gets or sets the name of the MySqlParameter.
		/// </summary>
		[Category("Misc")]
		public String ParameterName 
		{
			get { return paramName; }
			set 
			{ 
				paramName = value; 
				if (paramName[0] == '?' || paramName[0] == '@')
					paramName = paramName.Substring(1, paramName.Length-1);
			}
		}

		/// <summary>
		/// Gets or sets the maximum number of digits used to represent the <see cref="Value"/> property.
		/// </summary>
		[Category("Data")]
		public byte Precision 
		{
			get 
			{ 
				if (paramValue is MySqlDecimal)
					return (paramValue as MySqlDecimal).Precision;
				return 0;
			}
			set 
			{ 
				if (paramValue is MySqlDecimal)
					(paramValue as MySqlDecimal).Precision = value;
			}
		}

		/// <summary>
		/// Gets or sets the number of decimal places to which <see cref="Value"/> is resolved.
		/// </summary>
		[Category("Data")]
		public byte Scale 
		{
			get 
			{ 
				if (paramValue is MySqlDecimal)
					return (paramValue as MySqlDecimal).Scale;
				return 0;
			}
			set 
			{ 
				if (paramValue is MySqlDecimal)
					(paramValue as MySqlDecimal).Scale = value;
			}
		}

		/// <summary>
		/// Gets or sets the maximum size, in bytes, of the data within the column.
		/// </summary>
		[Category("Data")]
		public int Size 
		{
			get { return size; }
			set { size = value; }
		}

		/// <summary>
		/// Gets or sets the name of the source column that is mapped to the <see cref="DataSet"/> and used for loading or returning the <see cref="Value"/>.
		/// </summary>
		[Category("Data")]
		public String SourceColumn 
		{
			get { return sourceColumn; }
			set { sourceColumn = value; }
		}

		/// <summary>
		/// Gets or sets the <see cref="DataRowVersion"/> to use when loading <see cref="Value"/>.
		/// </summary>
		[Category("Data")]
		public DataRowVersion SourceVersion 
		{
			get { return sourceVersion; }
			set { sourceVersion = value; }
		}

		/// <summary>
		/// Gets or sets the value of the parameter.
		/// </summary>
		[TypeConverter(typeof(StringConverter))]
		[Category("Data")]
		public object Value 
		{
			get	{ return oValue; }
			set	{ oValue = value; }
		}

		#endregion

		private void EscapeByteArray( byte[] bytes, System.IO.MemoryStream s )
		{
			int theSize = size == 0 ? bytes.Length : size;
			s.Capacity += (int)((double)theSize * 1.5);

			for (int x=0; x < theSize; x++)
			{
				byte b = bytes[x];
				if (b == '\0') 
				{
					s.WriteByte( (byte)'\\' );
					s.WriteByte( (byte)'0' );
				}
				
				else if (b == '\\' || b == '\'' || b == '\"')
				{
					s.WriteByte( (byte)'\\' );
					s.WriteByte( b );
				}
				else
					s.WriteByte( b );
			}
		}

		/// <summary>
		/// Overridden. Gets a string containing the <see cref="ParameterName"/>.
		/// </summary>
		/// <returns></returns>
		public override string ToString() 
		{
			return paramName;
		}

		private string EscapeString( string s )
		{
			s = s.Replace("\\", "\\\\");
			s = s.Replace("\'", "\\\'");
			s = s.Replace("\"", "\\\"");
			s = s.Replace("`", "\\`");
			s = s.Replace("", "\\");
			return s;
		}

		internal void SerializeToBinary( PacketWriter writer ) 
		{
			if (oValue == null || oValue == DBNull.Value) return;
			if (paramValue == null)
				paramValue = MySqlValue.GetMySqlValueFromValue( oValue );
			paramValue.ValueAsObject = oValue;
			if (paramValue is MySqlString)
				(paramValue as MySqlString).SerializeBinary( writer );
			else
				paramValue.SerializeBinary( writer );
		}

		internal void SerializeToText( System.IO.MemoryStream s, MySqlConnection conn )
		{
			string valStr = String.Empty;

			if (oValue == null || oValue == DBNull.Value)
				valStr = "NULL";
			else 
			{
				if (paramValue == null)
					paramValue = MySqlValue.GetMySqlValueFromValue( oValue );
				paramValue.ValueAsObject = oValue;

				if (paramValue is MySqlBinary && (paramValue as MySqlBinary).IsBinary)
				{
					// if we are on 4.1 or later, then we need to tell the server to 
					// treat this data as binary
					if (conn.InternalConnection.Driver.Version.isAtLeast(4,1,0))
					{
						byte[] binarybuff = conn.Encoding.GetBytes( "_binary " );
						s.Write( binarybuff, 0, binarybuff.Length );
					}

					s.WriteByte((byte)'\'');
					EscapeByteArray( (byte[])paramValue.ValueAsObject, s );
					s.WriteByte((byte)'\'');
				}
				else 
				{
					if (paramValue is MySqlDateTime)
						valStr = (paramValue as MySqlDateTime).ToMySqlString(conn.InternalConnection.Driver.Version.isAtLeast(4,1,0));
					else 
						valStr = paramValue.ToString();

					if (paramValue is MySqlString || paramValue is MySqlDateTime ||
						paramValue is MySqlTimeSpan ||
						(paramValue is MySqlBinary && !(paramValue as MySqlBinary).IsBinary))
						valStr = "'" + EscapeString(valStr) + "'";
				}
			}

			byte[] bytes = conn.Encoding.GetBytes(valStr);
			s.Write(bytes, 0, bytes.Length);
		}


		#region ICloneable
		object System.ICloneable.Clone() 
		{
			MySqlParameter clone = new MySqlParameter( paramName, paramValue.MySqlDbType, direction,
				sourceColumn, sourceVersion, paramValue );
			return clone;
		}
		#endregion

		internal class MySqlParameterConverter : TypeConverter
		{
			public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
			{
				if (destinationType == typeof(InstanceDescriptor))
				{
					return true;
				}

				// Always call the base to see if it can perform the conversion.
				return base.CanConvertTo(context, destinationType);
			}

			public override object ConvertTo(ITypeDescriptorContext context, 
				System.Globalization.CultureInfo culture, object value, Type destinationType)
			{
				if (destinationType == typeof(InstanceDescriptor))
				{
					ConstructorInfo ci = typeof(MySqlParameter).GetConstructor(
						new Type[]{typeof(string), typeof(MySqlDbType), typeof(int), typeof(ParameterDirection),
									  typeof(bool), typeof(byte), typeof(byte), typeof(string), typeof(DataRowVersion),
									  typeof(object)});
					MySqlParameter p = (MySqlParameter) value;
					return new InstanceDescriptor(ci,new object[]{ 
																	 p.ParameterName, p.DbType, p.Size, p.Direction, p.IsNullable, p.Precision,
																	 p.Scale, p.SourceColumn, p.SourceVersion, p.Value});
				}

				// Always call base, even if you can't convert.
				return base.ConvertTo(context, culture, value, destinationType);
			}
		}
	}
}
