// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using System.Threading;
using MySql.Data.MySqlClient;
using NUnit.Framework;

namespace MySql.Data.MySqlClient.Tests
{
	/// <summary>
	/// Summary description for BlobTests.
	/// </summary>
	[NUnit.Framework.TestFixture]
	public class BlobTests : BaseTest
	{
		[TestFixtureSetUp]
		public void TestFixtureSetUp()
		{
			Open();

			execSQL("DROP TABLE IF EXISTS Test");
			execSQL("CREATE TABLE Test (id INT NOT NULL, blob1 LONGBLOB, text1 LONGTEXT, PRIMARY KEY(id))");
		}

		[TestFixtureTearDown]
		public void TestFixtureTearDown() 
		{
			Close();
		}

		[Test]
		public void InsertBinary() 
		{
			int lenIn = 400000;
			byte[] dataIn = Utils.CreateBlob(lenIn);

			MySqlCommand cmd = new MySqlCommand("TRUNCATE TABLE Test", conn);
			cmd.ExecuteNonQuery();

			cmd.CommandText = "INSERT INTO Test VALUES (?id, ?b1, NULL)";
			cmd.Parameters.Add( new MySqlParameter("?id", 1));
			cmd.Parameters.Add( new MySqlParameter("?b1", dataIn));
			int rows = cmd.ExecuteNonQuery();

			byte[] dataIn2 = Utils.CreateBlob(lenIn);
			cmd.Parameters[0].Value = 2;
			cmd.Parameters[1].Value = dataIn2;
			rows += cmd.ExecuteNonQuery();

			Assert.AreEqual( 2, rows, "Checking insert rowcount" );

			MySqlDataReader reader = null;
			try 
			{
				cmd.CommandText = "SELECT * FROM Test";
				reader = cmd.ExecuteReader();
				Assert.AreEqual( true, reader.HasRows, "Checking HasRows" );
			
				reader.Read();

				byte[] dataOut = new byte[ lenIn ];
				long lenOut = reader.GetBytes( 1, 0, dataOut, 0, lenIn );

				Assert.AreEqual( lenIn, lenOut, "Checking length of binary data (row 1)" );

				// now see if the buffer is intact
				for (int x=0; x < dataIn.Length; x++) 
					Assert.AreEqual( dataIn[x], dataOut[x], "Checking first binary array at " + x );

				reader.Read();
				lenOut = reader.GetBytes( 1, 0, dataOut, 0, lenIn );
				Assert.AreEqual( lenIn, lenOut, "Checking length of binary data (row 2)" );

				// now see if the buffer is intact
				for (int x=0; x < dataIn2.Length; x++) 
					Assert.AreEqual( dataIn2[x], dataOut[x], "Checking second binary array at " + x );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test]
		public void InsertText() 
		{
			byte[] data = new byte[1024];
			for (int x=0; x < 1024; x++) 
				data[x] = (byte)(65 + (x%20));

			// Create sample table
			MySqlCommand cmd = new MySqlCommand("TRUNCATE TABLE Test;INSERT INTO Test VALUES (1, NULL, ?t1)", conn);
			cmd.Parameters.Add( new MySqlParameter("?t1", data));
			int rows = cmd.ExecuteNonQuery();
			Assert.AreEqual( 1, rows, "Checking insert rowcount" );

			MySqlDataReader reader = null;
			try 
			{
				cmd.CommandText = "SELECT * FROM Test where id=1";
				reader = cmd.ExecuteReader();
				Assert.AreEqual( true, reader.HasRows, "Checking HasRows" );
			
				reader.Read();

				string s = reader.GetString(2);
				Assert.AreEqual( 1024, s.Length, "Checking length returned " );
				Assert.AreEqual( "ABCDEFGHI", s.Substring(0, 9), "Checking first few chars of string" );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test]
		public void UpdateDataSet() 
		{
			MySqlCommand cmd = new MySqlCommand("TRUNCATE TABLE Test; INSERT INTO Test VALUES( 1, NULL, NULL )", conn);
			cmd.ExecuteNonQuery();

			MySqlDataAdapter da = new MySqlDataAdapter("SELECT * FROM Test", conn);
			MySqlCommandBuilder cb = new MySqlCommandBuilder(da);
			DataTable dt = new DataTable();
			da.Fill(dt);

			byte[] inBuf = Utils.CreateBlob(512);
			dt.Rows[0]["blob1"] = inBuf;
			DataTable changes = dt.GetChanges();
			da.Update( changes );
			dt.AcceptChanges();

			dt.Clear();
			da.Fill(dt);

			byte[] outBuf = (byte[])dt.Rows[0]["blob1"];
			Assert.AreEqual( inBuf.Length, outBuf.Length, "checking length of updated buffer" );
			for (int y=0; y < inBuf.Length; y++)
				Assert.AreEqual( inBuf[y], outBuf[y], "checking array data" );
		}

	}
}
