// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Data;
using MySql.Data.MySqlClient;
using MySql.Data.Types;
using NUnit.Framework;

namespace MySql.Data.MySqlClient.Tests
{
	/// <summary>
	/// Summary description for ConnectionTests.
	/// </summary>
	[TestFixture] 
	public class DataReaderTests : BaseTest
	{
		[TestFixtureSetUp]
		public void TestFixtureSetUp()
		{
			Open();

			execSQL("DROP TABLE IF EXISTS Test");
			execSQL("CREATE TABLE Test (id INT NOT NULL, name VARCHAR(100), d DATE, dt DATETIME, b1 LONGBLOB, PRIMARY KEY(id))");
		}

		[TestFixtureTearDown]
		public void TestFixtureTearDown() 
		{
			Close();
		}

		[Test()]
		public void TestMultipleResultsets()
		{
			MySqlDataReader reader =null;
			try 
			{
				MySqlCommand cmd = new MySqlCommand("", conn);
				// insert 100 records
				cmd.CommandText = "INSERT INTO Test (id,name) VALUES (?id, 'test')";
				cmd.Parameters.Add( new MySqlParameter("?id", 1));
				for (int i=1; i <= 100; i++)
				{
					cmd.Parameters[0].Value = i;
					cmd.ExecuteNonQuery();
				}
				
				cmd = new MySqlCommand("SELECT id FROM test WHERE id<50; SELECT * FROM test WHERE id >= 50", conn);
				reader = cmd.ExecuteReader();
				Assert.IsNotNull( reader );
				Assert.AreEqual( true, reader.HasRows );
				Assert.IsTrue( reader.Read() );
				Assert.AreEqual( 1, reader.FieldCount );
				Assert.IsTrue( reader.NextResult() );
				Assert.AreEqual( true, reader.HasRows );
				Assert.AreEqual( 5, reader.FieldCount );
				reader.Close();

				cmd.CommandText = "DELETE FROM Test WHERE id>=0";
				cmd.ExecuteNonQuery();
			}
			catch (Exception ex)
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void TestNotReadingResultset()
		{
			for (int x=0; x < 10; x++)
			{
				MySqlConnection c = new MySqlConnection( conn.ConnectionString + ";pooling=false" );
				c.Open();

				MySqlCommand cmd = new MySqlCommand("INSERT INTO Test (id, name, b1) VALUES(?val, 'Test', NULL)", c);
				cmd.Parameters.Add( new MySqlParameter("?val", x));
				int affected = cmd.ExecuteNonQuery();
				Assert.AreEqual( 1, affected );

				cmd = new MySqlCommand("SELECT * FROM Test", c);
				cmd.ExecuteReader();
				c.Close();
			}
		}

		[Test()]
		public void GetBytes()
		{
			int len = 50000;
			byte[] bytes = Utils.CreateBlob( len );
			MySqlCommand cmd = new MySqlCommand("INSERT INTO Test (id, name, b1) VALUES(1, 'Test', ?b1)", conn);
			cmd.Parameters.Add( "?b1", bytes );
			cmd.ExecuteNonQuery();

			cmd.CommandText = "SELECT * FROM Test";
			MySqlDataReader reader = cmd.ExecuteReader();
			reader.Read();

			long sizeBytes = reader.GetBytes( 4, 0, null, 0, 0 );
			Assert.AreEqual( len, sizeBytes );

			byte[] buff1 = new byte[len/2];
			byte[] buff2 = new byte[len - (len/2)];
			long buff1cnt = reader.GetBytes( 4, 0, buff1, 0, len /2 );
			long buff2cnt = reader.GetBytes( 4, buff1cnt, buff2, 0, buff2.Length );
			Assert.AreEqual( buff1.Length, buff1cnt );
			Assert.AreEqual( buff2.Length, buff2cnt );

			for (int i=0; i<buff1.Length; i++)
				Assert.AreEqual( bytes[i], buff1[i] );

			for (int i=0; i<buff2.Length; i++)
				Assert.AreEqual( bytes[buff1.Length + i], buff2[i] );

			reader.Close();
		}

		[Test()]
		public void TestSingleResultSetBehavior()
		{
			execSQL("INSERT INTO Test (id, name, b1) VALUES (1, 'Test1', NULL)");
			execSQL("INSERT INTO Test (id, name, b1) VALUES (2, 'Test1', NULL)");

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test WHERE id=1; SELECT * FROM Test WHERE id=2", conn);
			MySqlDataReader reader = cmd.ExecuteReader( CommandBehavior.SingleResult );
			bool result = reader.Read();
			Assert.AreEqual( true, result );

			result = reader.NextResult();
			Assert.AreEqual( false, result );

			reader.Close();
		}

		[Test()]
		public void TestZeroDateTimeException() 
		{
			execSQL("INSERT INTO Test (id, d, dt) VALUES (1, '0000-00-00', '0000-00-00 00:00:00')");

			MySqlDataReader reader = null;
			try 
			{
				MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
				reader = cmd.ExecuteReader();
				reader.Read();
				DateTime dt = reader.GetDateTime(3);
				Assert.Fail("Should throw an exception");
			}
			catch (MySqlConversionException) 
			{
			}
			catch (MySqlException ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void TestRoundZeroDateTime() 
		{
			execSQL("INSERT INTO Test (id, d, dt) VALUES (1, '0000-00-00', '0000-00-00 00:00:00')");

			MySqlConnection c = new MySqlConnection(
				conn.ConnectionString + ";pooling=false;RoundZeroDatetime=true" );
			c.Open();
			MySqlDataReader reader = null;
			try 
			{
				MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", c);
				reader = cmd.ExecuteReader();
				reader.Read();
				DateTime dt = reader.GetDateTime(2);
				Assert.AreEqual( new DateTime(1,1,1,0,0,0), dt );
			}
			catch (MySqlException ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
				c.Close();
			}
		}

		[Test()]
		public void GetSchema() 
		{
			string sql = "CREATE TABLE test2( " +
				"id INT UNSIGNED AUTO_INCREMENT NOT NULL, " +
				"name VARCHAR(255) NOT NULL, " + 
				"PRIMARY KEY( id ))";

			execSQL("DROP TABLE IF EXISTS test2");
			execSQL(sql);
			execSQL("INSERT INTO test2 VALUES(1,'Test')");

			MySqlDataReader reader = null;

			try 
			{
				MySqlCommand cmd = new MySqlCommand("SELECT * FROM test2", conn);
				reader = cmd.ExecuteReader();
				DataTable dt = reader.GetSchemaTable();
				Assert.AreEqual( true, dt.Rows[0]["IsAutoIncrement"], "Checking auto increment" );
				Assert.AreEqual( true, dt.Rows[0]["IsUnique"], "Checking IsUnique" );
				Assert.AreEqual( false, dt.Rows[0]["AllowDBNull"], "Checking AllowDBNull" );
				Assert.AreEqual( false, dt.Rows[1]["AllowDBNull"], "Checking AllowDBNull" );
			}
			catch (Exception ex) 
			{
				Assert.Fail(ex.Message);
			}
			finally 
			{
				if (reader != null) reader.Close();
			}

			execSQL("DROP TABLE IF EXISTS test2");
		}

		[Test()]
		public void CloseConnectionBehavior() 
		{
			execSQL("INSERT INTO Test(id,name) VALUES(1,'test')");

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader( CommandBehavior.CloseConnection );
				Assert.IsTrue( reader.Read() );
				reader.Close();
				Assert.IsTrue( conn.State == ConnectionState.Closed );
				conn.Open();
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
				if (conn.State == ConnectionState.Closed) conn.Open();
			}
		}

		[Test()]
		public void SingleRowBehavior() 
		{
			execSQL("INSERT INTO Test(id,name) VALUES(1,'test1')");
			execSQL("INSERT INTO Test(id,name) VALUES(2,'test2')");
			execSQL("INSERT INTO Test(id,name) VALUES(3,'test3')");

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader( CommandBehavior.SingleRow );
				Assert.IsTrue( reader.Read(), "First read" );
				Assert.IsFalse( reader.Read(), "Second read" );
				Assert.IsFalse( reader.NextResult(), "Trying NextResult" );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void SingleRowBehaviorWithLimit() 
		{
			execSQL("INSERT INTO Test(id,name) VALUES(1,'test1')");
			execSQL("INSERT INTO Test(id,name) VALUES(2,'test2')");
			execSQL("INSERT INTO Test(id,name) VALUES(3,'test3')");

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test LIMIT 2", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader( CommandBehavior.SingleRow );
				Assert.IsTrue( reader.Read(), "First read" );
				Assert.IsFalse( reader.Read(), "Second read" );
				Assert.IsFalse( reader.NextResult(), "Trying NextResult" );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		//TODO: implement this
/*		[Test()]
		public void SequentialAccessBehavior() 
     	{
			execSQL("INSERT INTO Test(id,name) VALUES(1,'test1')");

			MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
			MySqlDataReader reader = null;
			try 
			{
				reader = cmd.ExecuteReader( CommandBehavior.SequentialAccess );
				Assert.IsTrue( reader.Read() );
				string s = reader.GetString( 1 );
				Assert.AreEqual( "test1", s );
				// this next line should throw an exception
				int i = reader.GetInt32( 0 );
				Assert.Fail( "This line should not execute" );
			}
			catch (MySqlException) { }
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}
*/
	}
}
