// Copyright (C) 2004 MySQL AB
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as published by
// the Free Software Foundation
//
// There are special exceptions to the terms and conditions of the GPL 
// as it is applied to this software. View the full text of the 
// exception in file EXCEPTIONS in the directory of this software 
// distribution.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using MySql.Data.MySqlClient;
using System.Data;
using NUnit.Framework;

namespace MySql.Data.MySqlClient.Tests
{
	/// <summary>
	/// Summary description for ConnectionTests.
	/// </summary>
	[TestFixture] 
	public class DataTypeTests : BaseTest
	{
		[TestFixtureSetUp]
		public void TestFixtureSetup()
		{
			Open();

			execSQL("DROP TABLE IF EXISTS Test");
			execSQL("CREATE TABLE Test (id INT NOT NULL, name VARCHAR(100), d DATE, dt DATETIME, tm TIME,  PRIMARY KEY(id))");
		}

		[TestFixtureTearDown()]
		public void TestFixtureTearDown() 
		{
			Close();
		}

		[Test]
		public void TestGuid()
		{
			MySqlDataReader reader =null;

			try 
			{
				MySqlCommand cmd = new MySqlCommand("TRUNCATE TABLE Test", conn);
				cmd.ExecuteNonQuery();

				Guid g = Guid.NewGuid();
				cmd.CommandText = "INSERT INTO Test VALUES (?id, ?guid, NULL, NULL, NULL)";
				cmd.Parameters.Add( new MySqlParameter("?id", 1));
				cmd.Parameters.Add( new MySqlParameter( "?guid", g ));
				cmd.ExecuteNonQuery();

				cmd.Parameters[0].Value = 2;
				cmd.Parameters[1].Value = g.ToString("N");
				cmd.ExecuteNonQuery();

				cmd.Parameters[0].Value = 3;
				cmd.Parameters[1].Value = g.ToString("D");
				cmd.ExecuteNonQuery();

				cmd.Parameters[0].Value = 4;
				cmd.Parameters[1].Value = g.ToString("B");
				cmd.ExecuteNonQuery();

				cmd.Parameters[0].Value = 5;
				cmd.Parameters[1].Value = g.ToString("P");
				cmd.ExecuteNonQuery();

				cmd.CommandText = "SELECT * FROM Test";
				reader = cmd.ExecuteReader();

				Assert.AreEqual( true, reader.Read() );
				Guid newG = reader.GetGuid(1);
				Assert.AreEqual( g, newG );

				Assert.AreEqual( true, reader.Read() );
				newG = reader.GetGuid(1);
				Assert.AreEqual( g, newG );

				Assert.AreEqual( true, reader.Read() );
				newG = reader.GetGuid(1);
				Assert.AreEqual( g, newG );

				Assert.AreEqual( true, reader.Read() );
				newG = reader.GetGuid(1);
				Assert.AreEqual( g, newG );
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void TestTime()
		{
			MySqlDataReader reader = null;

			try 
			{
				MySqlCommand cmd = new MySqlCommand("INSERT INTO TEST (id, tm) VALUES (1, '00:00')", conn);
				cmd.ExecuteNonQuery();
				cmd.CommandText = "INSERT INTO TEST (id, tm) VALUES (2, '512:45:17')";
				cmd.ExecuteNonQuery();

				cmd.CommandText = "SELECT * FROM TEST";
				reader = cmd.ExecuteReader();
				reader.Read();

				object value = reader["tm"];
				Assert.AreEqual( value.GetType(), typeof(TimeSpan));
				TimeSpan ts = (TimeSpan)reader["tm"];
				Assert.AreEqual( 0, ts.Hours );
				Assert.AreEqual( 0, ts.Minutes );
				Assert.AreEqual( 0, ts.Seconds );

				reader.Read();
				value = reader["tm"];
				Assert.AreEqual( value.GetType(), typeof(TimeSpan));
				ts = (TimeSpan)reader["tm"];
				Assert.AreEqual( 21, ts.Days );
				Assert.AreEqual( 8, ts.Hours );
				Assert.AreEqual( 45, ts.Minutes );
				Assert.AreEqual( 17, ts.Seconds );
			
				reader.Close();
			}
			catch (Exception ex)
			{
				Assert.Fail(ex.Message);
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}

		[Test()]
		public void TypeCoercion()
		{
			MySqlParameter p = new MySqlParameter("?test", 1);
			Assert.AreEqual( DbType.Int32, p.DbType );
			Assert.AreEqual( MySqlDbType.Int32, p.MySqlDbType );

			p.DbType = DbType.Int64;
			Assert.AreEqual( DbType.Int64, p.DbType );
			Assert.AreEqual( MySqlDbType.Int64, p.MySqlDbType );

			p.MySqlDbType = MySqlDbType.Int16;
			Assert.AreEqual( DbType.Int16, p.DbType );
			Assert.AreEqual( MySqlDbType.Int16, p.MySqlDbType );
		}

		[Test()]
		public void ZeroDateAndTime() 
		{
			execSQL("INSERT INTO Test VALUES(1, 'Test', '0000-00-00', '0000-00-00', '00:00:00')");

			MySqlDataReader reader = null;
			try 
			{
				MySqlCommand cmd = new MySqlCommand("SELECT * FROM Test", conn);
				reader = cmd.ExecuteReader();
				while (reader.Read()) 
				{
					Assert.AreEqual( "0000-00-00", reader.GetValue(2).ToString() );
					Assert.AreEqual( "0000-00-00 00:00:00", reader.GetValue(3).ToString() );
					Assert.AreEqual( "00:00:00", reader.GetValue(4).ToString() );
				}
			}
			catch (Exception ex) 
			{
				Assert.Fail( ex.Message );
			}
			finally 
			{
				if (reader != null) reader.Close();
			}
		}
	}
}
